import asyncio
import sys

from typing import Any, Generator, Generic, TypeVar, Optional

if sys.version_info >= (3, 9):
    from collections.abc import Iterable, Iterator
    List = list  # typing.List deprecated, use builtins.list instead
else:
    from typing import Iterable, Iterator, List


__all__ = (
    'AsyncIterable',
    'async_iter'
)


T = TypeVar('T')


class AsyncIterable(Generic[T]):
    def __init__(self, iterable: Iterable[T], *, delay: float = 0):
        self.iterable = iterable
        self.iterator: Optional[Iterator[T]] = None

        self.delay = delay

    def __await__(self) -> Generator[Any, Any, List[T]]:
        return self.flatten().__await__()

    def __aiter__(self):
        self.iterator = iter(self.iterable)
        return self

    async def __anext__(self) -> T:
        assert self.iterator is not None

        try:
            item: T = next(self.iterator)
        except StopIteration:
            self.iterator = None
            raise StopAsyncIteration

        await asyncio.sleep(self.delay)
        return item

    async def flatten(self) -> List[T]:
        return [item async for item in self]


def async_iter(iterable: Iterable[T], *, delay: float = 0) -> AsyncIterable[T]:
    """
    Asynchronously iterate with an optional delay in between iterations.

    Parameters
    -----------
    iterable: :class:`Iterable`
        Anything that can be synchronously be iterated over.
    delay: Optional[:class:`float`]
        An optional delay between each iteration.

    Example
    --------
    .. code:: py

        import asyncify

        async def main():
            async for num in asyncify.async_iter([1, 2, 3], delay=2):
                print('%d seconds have passed!' % (num * 2))
    """
    return AsyncIterable(iterable, delay=delay)



import inspect
import sys
import types
from typing import TypeVar

from .func import asyncify_func

if sys.version_info >= (3, 10):
    from collections.abc import Callable
else:
    from typing import Callable


__all__ = (
    'asyncify_class',
    'ignore'
)


CallableT = TypeVar('CallableT', bound=Callable)
TypeT = TypeVar('TypeT', bound=type)


function_types: tuple[type, ...] = (
    types.FunctionType,
    classmethod,
    staticmethod
)


def ignore(func: CallableT) -> CallableT:
    """
    A decorator to ignore a function in a class when using :func:`asyncify.asyncify_class`.
    """
    func._asyncify_ignore = True  # type: ignore
    return func


def asyncify_class(cls: TypeT) -> TypeT:
    """
    Turn a classes methods into async functions.
    This uses :func:`asyncify.asyncify_func`.
    This ignores methods marked with :func:`asyncify.ignore` and `dunder` methods.

    Example
    ---------
    .. code:: py

        import asyncify
        import requests

        @asyncify.asyncify_class
        class RequestsClient:
            def __init__(self):  # ignored by asyncify
                self.session = requests.Session()

            def request(self, method: str, url: str):  # now a coroutine function
                return self.session.request(method, url)

        # can also be used like this
        RequestsClient = asyncify.asyncify_class(requests.Session)

        client = RequestsClient()

        async def main():
            await client.request('GET', 'https://python.org')
    """

    for name, func in inspect.getmembers(cls):
        if not isinstance(
                func, function_types
        ) or getattr(
            func, '_asyncify_ignore', False
        ) or name.startswith('__'):
            continue

        func = asyncify_func(func)
        setattr(cls, name, func)

    return cls
